@@ -1,6 +1,6 @@
 # Almost Free Text (AFT) Parser
 #
-# Copyright (C) 1996-2005 Todd A. Coram. All rights reserved.
+# Copyright (C) 1996-2009 Todd A. Coram. All rights reserved.
 #
 # This perl script parses aft documents and produces output formatted according
 # to an aft 'element' file.  See aft-refman.aft  for additional information.
@@ -13,16 +13,16 @@ package AFT;
 
 use English;
 use vars qw ($VERSION $version $outputfile $element_type $author $title
-	     $pre_process_line);
+	     $pre_process_line $my_print);
 
 # Initializations of globals that consumers of this package may read/modify.
 #
 
 $author = '';
 $title = '';
-$VERSION="v5.096";
+$VERSION="v5.097";
 $version=
-  "Almost Free Text $VERSION; Copyright 1996-2005 Todd Coram. All rights reserved.";
+  "Almost Free Text $VERSION; Copyright 1996-2009 Todd Coram.";
 
 $element_type="bn-html";		# Default
 $outputfile='';			# Output file
@@ -31,6 +31,9 @@ $outputfile='';			# Output file
 # This is of interest to apps that use AFT.pm as a package. 
 #
 $pre_process_line = sub { local $_ = shift; my($fname,$lcnt) = @_; return $_; };
+# You can supply your own print subroutine here.
+#
+$my_print = sub { print @_; };
 
 # Package scoped variables.
 #
@@ -73,9 +76,9 @@ my @list_stack;			# A stack of lists as we nest.
 my @section_stack;		# Keeps track of nesting sections.
 my $section_number = Autonum->new();
 
-my @index;			# Holds running index targets. TBD
+my %index;			# Hash of arrays (name -> [ ref, ... ])
 
-my @note;			# Holds collected 'endnotes'. TBD
+my @note;			# Holds collected 'endnotes'.
 
 
 #
@@ -103,6 +106,15 @@ my $face =
 my %pragma_prevar = ();		# variables expanded before filtering.
 my %pragma_ctl = ();		# variables used for internal control
 
+# Set up some convenient symbols %%
+#
+sub setup_symbols {
+  $pragma_prevar{'lb'} = $AFT_OUTPUT::elem{'LineBreak'};
+  $pragma_prevar{'sp'} = $AFT_OUTPUT::elem{'NBSPACE'};
+  $pragma_prevar{'bang'} = '!';
+}
+
+
 # Don't expand pragmas in verbatim mode.
 #
 $AFT_OUTPUT::pragma_ctl{expandinverbatim} = 'no';
@@ -139,6 +151,7 @@ my @Functions = (
 sub main {
     &parse_command_line();
     &load_element_file(&element_file());
+    &setup_symbols();
 
     if ($outputfile eq '') {
         # Use first input filename to construct output filename.
@@ -247,11 +260,11 @@ sub do_string {
 
 # Run AFT on a file given its file handle.
 #
-sub do_f_h__file {
+sub do_FH_file {
     my($fh, $filename) = @_;
     &load_element_file(&element_file());
     &begin();
-    &process_f_h__file($fh, $filename);
+    &process_FH_file($fh, $filename);
     &end();
 }
 
@@ -265,11 +278,26 @@ sub output_postamble {
     &output($AFT_OUTPUT::file_postamble."\n", aft => &filter($aft_advert));
 }
 
+sub output_indices {
+  return if (%index == 0);
+  &output($AFT_OUTPUT::elem{'HorizontalLine'}."\n");
+  &do_string("* Index");
+  &do_string("\n");
+  foreach my $key (sort(keys %index)) {
+    next if ($key =~ /iNtErNaLNOTE/);
+    &output($AFT_OUTPUT::elem{'LineBreak'});
+    &output("${key} : ");
+    foreach my $target (@{$index{$key}}) {
+      &do_string("[*($target)], ");
+    }
+  }
+}
+
 sub output_notes {
     &output($AFT_OUTPUT::elem{'HorizontalLine'}."\n");
     my $count = 1;
     foreach my $note (@note) {
-	&do_string("=[(fNOTE$count)]=\n\\[[$count(REFfNOTE$count)]] - $note\n");
+	&do_string("=[(iNtErNaLNOTE$count)]=\n\\[[$count(REFiNtErNaLNOTE$count)]] - $note\n");
 	&output($AFT_OUTPUT::elem{'LineBreak'});
 	$count++;
     }
@@ -286,6 +314,7 @@ sub end {
 	if ($mode->{haveSections});
     
     &output_notes() if ($AFT_OUTPUT::elem{'NotesAtEnd?'} eq 'yes');
+#    &output_indices();
 
     # End output file with Postamble..
     #
@@ -296,18 +325,17 @@ sub parse_command_line {
     ## Process the command line options.
     #
     my $usage=
-	"Usage:\n aft [--autonumber] [--verbose] [--output=file | --output=-] ".
-	"[--type=output_type] infile ..";
-    
+	"Usage:\n aft [--autonumber] [--verbose] [--output=file | --output=-]".
+	" [--type=output_type] infile ..";
+
     use Getopt::Long;
     GetOptions ("output=s" => \$outputfile, # output file name
 		"verbose!" => \$verbose, # output type (html, etc) 
 		"type=s" => \$element_type, # output type (html, etc)
 		"autonumber!" => \$autonumber, # section numbers
 		"tabstop=i" => \$tabstop); # number of spaces = tab
-    
+
     print (STDERR "$version\n$usage\n"), exit 2 if (@ARGV == 0);
-    
 }
 
 sub element_file {
@@ -343,7 +371,7 @@ sub process_file {
     # Do that voodoo that you do so well.
     #
     print (STDERR "\nProcessing $fname.\n[") if $verbose;
-    &process_f_h__file(*IN, $fname);
+    &process_FH_file(*IN, $fname);
     
     # Done with it, so close it.
     #
@@ -355,14 +383,15 @@ sub process_file {
 
 # processFH_File (fh,fname) - Process the supplied file by the handle.
 #
-sub process_f_h__file {
+sub process_FH_file {
     my($fh, $fname) = @_;
     my $lcnt  = 0;		# line count
     
     my $continued_line = "";
     LINE: while (<$fh>) {
 	$lcnt++;
-	chop;
+	chomp;
+	chop if /.*\r$/;  # In case we are unix perl processing a MSDOS .aft file
 	$continued_line .= $1, next LINE if (/(.*)\\$/ and !$mode->{inVerbatim}); # collect continuations
 	&do_string($continued_line.$_, $fname, $lcnt); # process complete line
 	$continued_line = "";
@@ -520,13 +549,13 @@ sub handle_ruler {
     return 0;
 }
 
-# Handle *, **, ***, **** (sections) and
+# Handle *, **, ***, ****, **** etc (sections) and
 # ^*, ^**, ^***, ^**** (sections referencing TOC)
 #
 sub handle_sections {
     my $fname = shift; my $lcnt = shift; local($_) = @_;
     
-    /^(\^*\*{1,4})\s*(.+?)\s*$/ and do {
+    /^(\^*\*{1,7})\s*(.+?)\s*$/ and do {
 	my($sname) = $2;
 	if ($mode->{haveSections} eq 0) {
 	    &output($AFT_OUTPUT::elem{'BeginSectLevel1'}."\n");
@@ -548,10 +577,12 @@ sub handle_sections {
 	
 	# Save the section for the TOC file.
 	#
-	my($level) = $1;
-	$level =~ tr/*^/\t/d;
+	if (length($1) < 5) {
+	  my($level) = $1;
+	  $level =~ tr/*^/\t/d;
 	
-	print ($tocout "$level"."* {-$full_sname\@$sname-}\n") if $tocout;
+	  print ($tocout "$level"."* {-$full_sname\@$sname-}\n") if $tocout;
+	}
 	return 1;
     };
     return 0;
@@ -790,6 +821,9 @@ sub handle_new_table_parser {
 	
 	my $tline = $1;
 
+	# Protect \!
+	$tline =~ s/\\!/%bang%/g;
+
 	# First thing is first... Are we in the table yet?
 	#
 	!$mode->{inTable} and do {
@@ -801,7 +835,8 @@ sub handle_new_table_parser {
 	    # time through.
 	    #
 	    $mode->{needTableHeaders} = 1;
-	    if ($1 =~ /([^\!]*)/) { $tline = $1; }
+	    if ($tline =~ /([^\!]*)/) { $tline = $1; }
+#	    if ($1 =~ /([^\!]*)/) { $tline = $1; }
 	    # The first thing we got was a caption. Save it for later.
 	    #
 	    $table_caption = &filter($tline);
@@ -841,6 +876,7 @@ sub handle_new_table_parser {
 	# trailing column. If we got less than 2 elements, this ain't no
 	# table!
 	#
+
 	if (($ecnt = (@elements = split ("!", $tline, 100)) - 1) < 2) {
 	    print(STDERR 
 		   "\n$fname($lcnt): Weirdness in a table... not enough columns.\n");
@@ -970,7 +1006,7 @@ sub generate_t_oC {
 	# Read it in.
 	#
 	print (STDERR "\t Reading table of contents from $tocfile...")if $verbose;
-	&process_f_h__file(*TOCIN, $tocfile);
+	&process_FH_file(*TOCIN, $tocfile);
 	close(TOCIN);
 	&output("\n\n");
 	print (STDERR "Done.\n") if $verbose;
@@ -1068,11 +1104,11 @@ sub handle_notes {
     # look for (and replace) [Note: .. ]
     #
     $line =~ s/\[Note:\s*(.*)\]/
-	&handle_links("=[(REFfNOTE".$notenum.")]=").
+	&handle_links("=[(REFiNtErNaLNOTE".$notenum.")]=").
 	&prepare_output($AFT_OUTPUT::elem{'Note'}, 
 		       note => "$1",
 		       notereftxt => $notenum,
-		       notetarget => "fNOTE$notenum")/eg
+		       notetarget => "iNtErNaLNOTE$notenum")/eg
 		       and push(@note, $1); # save note
     return $line;
 }
@@ -1083,20 +1119,28 @@ sub handle_links {
     my($line) = @_;
     
     if ($AFT_OUTPUT::pragma_ctl{verbatimsquarebrackets} eq 'no') {
-	# new =[(target)]= style
+	# =[(target)]=
 	#
 	$line =~ s/=\[\(([^\[]+?)\)\]=/
 	    &prepare_output($AFT_OUTPUT::elem{'Target'},
 			   target => "$1", 
-			   text => $AFT_OUTPUT::elem{'NBSPACE'})/eg
-			   and push(@index, $1); # save for index
+			   text => $AFT_OUTPUT::elem{'NBSPACE'})/eg;
 	
-	# new =[target]= style
+	# =[^indexed]= - Generate a unique target (by appending array length).
+	#
+	my $_cnt;
+	$line =~ s/=\[\^([^\]]+?)\]=/
+	  $_cnt=@{$index{$1}},
+	  push(@{$index{$1}},"$1$_cnt"),
+	    &prepare_output($AFT_OUTPUT::elem{'Target'},
+			    target => "$1$_cnt", text => "$1")/eg;
+	# =[target]=
 	#
 	$line =~ s/=\[([^\]]+?)\]=/
 	    &prepare_output($AFT_OUTPUT::elem{'Target'},
-			   target => "$1", text => "$1")/eg
-			   and push(@index, $1); # save for index
+			    target => "$1", text => "$1")/eg;
+
+				 
 	
 	# new [name (url:reference)] style
 	#
@@ -1320,7 +1364,7 @@ sub set_prevar {
 # %key% is replaced by value everywhere in text.
 #
 sub output {
-    print &prepare_output(@_);
+    $my_print->(&prepare_output(@_));
 }
 
 # Prepare a line of text for output.
